//
// Software License for AMDiS
//
// Copyright (c) 2010 Dresden University of Technology 
// All rights reserved.
// Authors: Simon Vey, Thomas Witkowski et al.
//
// This file is part of AMDiS
//
// See also license.opensource.txt in the distribution.


#include "parallel/FeSpaceMapping.h"
#include "parallel/StdMpi.h"

namespace AMDiS {

  using namespace std;

  void GlobalDofMap::clear()
  {
    dofMap.clear();
    nRankDofs = 0;
    nOverallDofs = 0;
    rStartDofs = 0;
  }


  void GlobalDofMap::update()
  {
    for (map<DegreeOfFreedom, MultiIndex>::iterator it = dofMap.begin(); it != dofMap.end(); ++it)
      if (it->second.local == -1 && nonRankDofs.count(it->first) == 0)
	it->second.local = nRankDofs++;
    
    nOverallDofs = 0;
    rStartDofs = 0;
    mpi::getDofNumbering(*mpiComm, nRankDofs, rStartDofs, nOverallDofs);
    
    if (needGlobalMapping)
      computeGlobalMapping(rStartDofs);

    if (hasNonLocalDofs)
      computeNonLocalIndices();
  }


  void GlobalDofMap::computeGlobalMapping(int offset)
  {
    for (map<DegreeOfFreedom, MultiIndex>::iterator it = dofMap.begin(); it != dofMap.end(); ++it)
      it->second.global = it->second.local + offset;
  }


  void GlobalDofMap::computeNonLocalIndices()
  {
    FUNCNAME("GlobalDofMap::computeNonLocalIndices()");

    typedef map<int, map<const FiniteElemSpace*, DofContainer> >::iterator it_type;

    StdMpi<vector<int> > stdMpi(*mpiComm);
    for (DofComm::Iterator it(*sendDofs, feSpace); !it.end(); it.nextRank())
      for (; !it.endDofIter(); it.nextDof())
	if (dofMap.count(it.getDofIndex()) && !nonRankDofs.count(it.getDofIndex()))
	  stdMpi.getSendData(it.getRank()).push_back(dofMap[it.getDofIndex()].global);

    stdMpi.updateSendDataSize();

    for (DofComm::Iterator it(*recvDofs, feSpace); !it.end(); it.nextRank()) {
      bool recvFromRank = false;
      for (; !it.endDofIter(); it.nextDof()) {
	if (nonRankDofs.count(it.getDofIndex())) {
	  recvFromRank = true;
	  break;
	}
      }

      if (recvFromRank) 
	stdMpi.recv(it.getRank());
    }

    stdMpi.startCommunication();

    for (DofComm::Iterator it(*recvDofs, feSpace);
	 !it.end(); it.nextRank()) {
      int i = 0;
      for (; !it.endDofIter(); it.nextDof())
	if (nonRankDofs.count(it.getDofIndex()))
	  dofMap[it.getDofIndex()].global = stdMpi.getRecvData(it.getRank())[i++];
    }
  }


  void GlobalDofMap::print()
  {
    FUNCNAME("GlobalDofMap::print()");
    
    MSG("Local to global mapping on this rank: \n");

    for (map<DegreeOfFreedom, MultiIndex>::iterator it = dofMap.begin(); it != dofMap.end(); ++it)
      if (nonRankDofs.count(it->first) == 0)	
	MSG("  %d -> %d  (rank-dof)\n", it->first, it->second.local);
      else
	MSG("  %d -> %d  \n", it->first, it->second.local);
  }
}
